part of datagrid;

class _ColumnSizer {
  _ColumnSizer() {
    _isColumnSizerLoadedInitially = false;
  }

  GridColumn? _autoFillColumn;
  bool _isColumnSizerLoadedInitially = false;

  static const double _sortIconWidth = 20.0;
  static const double _sortNumberWidth = 18.0;

  late _DataGridStateDetails _dataGridStateDetails;

  void _initialRefresh(double availableWidth) {
    final _LineSizeCollection lineSizeCollection =
        _dataGridStateDetails().container.columnWidths as _LineSizeCollection;
    lineSizeCollection.suspendUpdates();
    _refresh(availableWidth);
    lineSizeCollection.resumeUpdates();
  }

  void _refresh(double availableWidth) {
    final _DataGridSettings dataGridSettings = _dataGridStateDetails();
    final bool hasAnySizerColumn = dataGridSettings.columns.any(
        (GridColumn column) =>
            (column.columnWidthMode != ColumnWidthMode.none) ||
            (column.width != double.nan) ||
            !column.visible);

    final _PaddedEditableLineSizeHostBase paddedEditableLineSizeHostBase =
        dataGridSettings.container.columnWidths;
    final _LineSizeCollection? lineSizeCollection =
        paddedEditableLineSizeHostBase is _LineSizeCollection
            ? paddedEditableLineSizeHostBase
            : null;

    if (lineSizeCollection == null) {
      return;
    }

    lineSizeCollection.suspendUpdates();
    _ensureColumnVisibility();

    if (dataGridSettings.columnWidthMode != ColumnWidthMode.none ||
        hasAnySizerColumn) {
      _sizerColumnWidth(availableWidth);
    }
    dataGridSettings.container.updateScrollBars();
    lineSizeCollection.resumeUpdates();
  }

  void _ensureColumnVisibility() {
    final _DataGridSettings dataGridSettings = _dataGridStateDetails();
    for (final GridColumn column in dataGridSettings.columns) {
      final int index = _GridIndexResolver.resolveToScrollColumnIndex(
          dataGridSettings, dataGridSettings.columns.indexOf(column));
      if (column.visible) {
        dataGridSettings.container.columnWidths.setHidden(index, index, false);
      } else {
        dataGridSettings.container.columnWidths.setHidden(index, index, true);
      }
    }
    // Columns will be auto sized only if Columns doesn't have explicit width
    // defined.
    _sizerColumnWidth(0.0);
  }

  void _sizerColumnWidth(double viewPortWidth) {
    final _DataGridSettings dataGridSettings = _dataGridStateDetails();
    double totalColumnSize = 0.0;
    final List<GridColumn> calculatedColumns = <GridColumn>[];

    _autoFillColumn = _getColumnToFill();

    // Hide Hidden columns
    final List<GridColumn> hiddenColumns = dataGridSettings.columns
        .where((GridColumn column) => !column.visible)
        .toList();
    for (final GridColumn column in hiddenColumns) {
      final int index = _GridIndexResolver.resolveToScrollColumnIndex(
          dataGridSettings, dataGridSettings.columns.indexOf(column));
      dataGridSettings.container.columnWidths.setHidden(index, index, true);
      calculatedColumns.add(column);
    }

    // Set width based on Column.Width
    final List<GridColumn> widthColumns = dataGridSettings.columns
        .skipWhile((GridColumn column) => !column.visible)
        .where((GridColumn column) => !(column.width).isNaN)
        .toList();
    for (final GridColumn column in widthColumns) {
      totalColumnSize += _setColumnWidth(column, column.width);
      calculatedColumns.add(column);
    }

    final List<GridColumn> lastColumnFill = dataGridSettings.columns
        .skipWhile((GridColumn column) => calculatedColumns.contains(column))
        .where((GridColumn col) =>
            (col.columnWidthMode == ColumnWidthMode.lastColumnFill) &&
            !_isLastFillColum(col))
        .toList();

    final List<GridColumn> autoColumns = lastColumnFill;

    for (final GridColumn column in autoColumns) {
      if ((column._autoWidth).isNaN) {
        final double columnWidth =
            dataGridSettings.container.columnWidths.defaultLineSize;
        totalColumnSize += columnWidth;
        _setAutoWidth(column, columnWidth);
      } else {
        totalColumnSize += _setColumnWidth(column, column._autoWidth);
      }
      calculatedColumns.add(column);
    }
    _setWidthBasedOnGrid(totalColumnSize, calculatedColumns, viewPortWidth);
    _autoFillColumn = null;
  }

  GridColumn? _getColumnToFill() {
    final _DataGridSettings dataGridSettings = _dataGridStateDetails();
    final GridColumn? column = dataGridSettings.columns.lastWhereOrNull(
        (GridColumn c) =>
            c.visible &&
            (c.width).isNaN &&
            (c.columnWidthMode == ColumnWidthMode.lastColumnFill));
    if (column != null) {
      return column;
    } else {
      if (dataGridSettings.columnWidthMode == ColumnWidthMode.lastColumnFill) {
        final GridColumn? lastColumn = dataGridSettings.columns
            .lastWhereOrNull((GridColumn c) => c.visible && (c.width).isNaN);
        if (lastColumn == null) {
          return null;
        }

        if (lastColumn.columnWidthMode == ColumnWidthMode.none) {
          return lastColumn;
        }
      }
    }
    return null;
  }

  void _setWidthBasedOnGrid(double totalColumnSize,
      List<GridColumn> calculatedColumns, double viewPortWidth) {
    final _DataGridSettings dataGridSettings = _dataGridStateDetails();
    for (final GridColumn column in dataGridSettings.columns) {
      if (calculatedColumns.contains(column)) {
        continue;
      }

      if (column.columnWidthMode == ColumnWidthMode.fill ||
          _isLastFillColum(column)) {
        continue;
      }

      switch (dataGridSettings.columnWidthMode) {
        case ColumnWidthMode.lastColumnFill:
          if ((column.width).isNaN) {
            totalColumnSize += _setColumnWidth(column,
                dataGridSettings.container.columnWidths.defaultLineSize);
            calculatedColumns.add(column);
          } else {
            totalColumnSize += _setColumnWidth(column, column.width);
            calculatedColumns.add(column);
          }
          break;
        case ColumnWidthMode.none:
          if (column.visible) {
            totalColumnSize += _setColumnWidth(column,
                dataGridSettings.container.columnWidths.defaultLineSize);
            calculatedColumns.add(column);
          }
          break;
        default:
          break;
      }
    }

    final List<GridColumn> remainingColumns = <GridColumn>[];

    for (final GridColumn column in dataGridSettings.columns) {
      if (!calculatedColumns.contains(column)) {
        remainingColumns.add(column);
      }
    }

    final double remainingColumnWidths = viewPortWidth - totalColumnSize;

    if (remainingColumnWidths > 0 &&
        (totalColumnSize != 0 ||
            (totalColumnSize == 0 && remainingColumns.length == 1) ||
            (dataGridSettings.columns.any((GridColumn col) =>
                    col.columnWidthMode == ColumnWidthMode.fill) ||
                dataGridSettings.columnWidthMode == ColumnWidthMode.fill))) {
      _setFillWidth(remainingColumnWidths, remainingColumns);
    } else {
      _setRemainingColumnsWidth(remainingColumns);
    }
  }

  void _setFillWidth(
      double remainingColumnWidth, List<GridColumn> remainingColumns) {
    final List<GridColumn> removedColumns = <GridColumn>[];
    final List<GridColumn> columns = remainingColumns.toList();
    double totalRemainingFillValue = remainingColumnWidth;

    double removedWidth = 0;
    GridColumn? fillColumn;
    bool isRemoved;
    while (columns.isNotEmpty) {
      isRemoved = false;
      removedWidth = 0;
      final double fillWidth =
          (totalRemainingFillValue / columns.length).floorToDouble();
      final GridColumn column = columns.first;
      if (column == _autoFillColumn &&
          (column.columnWidthMode == ColumnWidthMode.lastColumnFill ||
              (_dataGridStateDetails().columnWidthMode ==
                  ColumnWidthMode.lastColumnFill))) {
        columns.remove(column);
        fillColumn = column;
        continue;
      }
      final double computedWidth = _setColumnWidth(column, fillWidth);
      if (fillWidth != computedWidth && fillWidth > 0) {
        isRemoved = true;
        columns.remove(column);
        for (final GridColumn removedColumn in removedColumns) {
          if (!columns.contains(removedColumn)) {
            removedWidth += removedColumn._actualWidth;
            columns.add(removedColumn);
          }
        }
        removedColumns.clear();
        totalRemainingFillValue += removedWidth;
      }
      column._actualWidth = computedWidth;
      totalRemainingFillValue = totalRemainingFillValue - computedWidth;
      if (!isRemoved) {
        columns.remove(column);
        if (!removedColumns.contains(column)) {
          removedColumns.add(column);
        }
      }
    }

    if (fillColumn != null) {
      double columnWidth = 0;
      if ((fillColumn._autoWidth).isNaN) {
        columnWidth = 0.0;
        _setAutoWidth(fillColumn, columnWidth);
      } else {
        columnWidth = fillColumn._autoWidth;
      }

      if (totalRemainingFillValue < columnWidth) {
        _setColumnWidth(fillColumn, columnWidth);
      } else {
        _setColumnWidth(fillColumn, totalRemainingFillValue);
      }
    }
  }

  void _setRemainingColumnsWidth(List<GridColumn> remainingColumns) {
    final _DataGridSettings dataGridSettings = _dataGridStateDetails();
    for (final GridColumn column in remainingColumns) {
      if (_isLastFillColum(column)) {
        _setNoneWidth(
            column, dataGridSettings.container.columnWidths.defaultLineSize);
      } else if (!_isFillColumn(column)) {
        _setNoneWidth(
            column, dataGridSettings.container.columnWidths.defaultLineSize);
      }
    }
  }

  bool _isFillColumn(GridColumn column) {
    if (!column.width.isNaN) {
      return false;
    } else {
      return column.columnWidthMode == ColumnWidthMode.none
          ? _dataGridStateDetails().columnWidthMode == ColumnWidthMode.fill
          : column.columnWidthMode == ColumnWidthMode.fill;
    }
  }

  double _setNoneWidth(GridColumn column, double width) =>
      _setColumnWidth(column, width);

  bool _isLastFillColum(GridColumn column) {
    if (column == _autoFillColumn) {
      return true;
    }

    return false;
  }

  void _setAutoWidth(GridColumn? column, double width) {
    if (column != null) {
      column._autoWidth = width;
    }
  }

  void _resetAutoCalculation() {
    final _DataGridSettings dataGridSettings = _dataGridStateDetails();
    for (final GridColumn column in dataGridSettings.columns) {
      column._autoWidth = double.nan;
    }
  }

  double _getSortIconWidth(GridColumn column) {
    final _DataGridSettings dataGridSettings = _dataGridStateDetails();
    double width = 0.0;
    if (column.allowSorting && dataGridSettings.allowSorting) {
      width += _sortIconWidth;
      if (dataGridSettings.allowMultiColumnSorting &&
          dataGridSettings.showSortNumbers) {
        width += _sortNumberWidth;
      }
    }
    return width;
  }

  double _setColumnWidth(GridColumn column, double columnWidth) {
    final _DataGridSettings dataGridSettings = _dataGridStateDetails();
    final int columnIndex = dataGridSettings.columns.indexOf(column);
    final double width = _getColumnWidth(column, columnWidth);
    column._actualWidth = width;
    dataGridSettings.container.columnWidths[columnIndex] = column._actualWidth;
    return width;
  }

  double _getColumnWidth(GridColumn column, double columnWidth) {
    final _DataGridSettings dataGridSettings = _dataGridStateDetails();
    final int columnIndex = dataGridSettings.columns.indexOf(column);
    if (column.width < column._actualWidth) {
      return columnWidth;
    }

    final double width = dataGridSettings.container.columnWidths[columnIndex];

    final double resultWidth =
        _checkWidthConstraints(column, columnWidth, width);
    return resultWidth;
  }

  double _checkWidthConstraints(
      GridColumn column, double width, double columnWidth) {
    if (!(column.minimumWidth).isNaN || !(column.maximumWidth).isNaN) {
      if (!(column.maximumWidth).isNaN) {
        if (!width.isNaN && column.maximumWidth > width) {
          columnWidth = width;
        } else {
          columnWidth = column.maximumWidth;
        }
      }

      if (!(column.minimumWidth).isNaN) {
        if (!width.isNaN && column.minimumWidth < width) {
          if (width > column.maximumWidth) {
            columnWidth = column.maximumWidth;
          } else {
            columnWidth = width;
          }
        } else {
          columnWidth = column.minimumWidth;
        }
      }
    } else {
      if (!width.isNaN) {
        columnWidth = width;
      }
    }
    return columnWidth;
  }
}
